/************************************************************************************************************
* isolation -f /path/demfile [-d 0] [-r 100000]                                                             *
*                                                                                                           *
* Calculates topographic isolation of peaks based on elevation data                                         *
* Doc: https://wiki.openstreetmap.org/wiki/User:Maxbe/Dominanz_von_Gipfeln                                  *
*                                                                                                           *
* For installation you need the GDAL-libraries (debian: package libgdal-dev)                                *
* Compiling with "cc -Wall -o isolationOSMcomp2SRTM isolation.c -lgdal -lm -O2"                             *
*                                                                                                           *
************************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include "gdal/gdal.h"
#include "gdal/cpl_conv.h"
#include "gdal/gdal_frmts.h"

struct list_peak{
 long long int    id;
 long long int    heigherpeak_id;
 long int 		  higherpoint_ele;
 double           heigherpoint_lon,heigherpoint_lat;
 double           lon,lat,ele,eleDEM,eleOSM,isolation,quot;
 char             elestring[100];								 
};    
  
   

#define PI             3.1415927
/* min isolation of a peak */
#define MINISO         100
/* difference between the heights of DEM and peak to define a isolation */
#define MINDIFF        40
/* progress indicator modulus */
#define MODDEM         1000
#define MODINP         5000
#define OSMNOELE       -500000.0

int	debuglevel=0;
time_t starttime;
int	convertFt2Meter=0;
long miniso=MINISO, mindiff=MINDIFF;

void printhelp(){
 fprintf(stderr,"Check stderr\n");
 printf("Usage:   isolation -f demfile [options]\n");
 printf("         -f demfile\n         -d debuglevel (0-4, default=0)\n");
 printf("         -n maximal number of peaks (default 2000000)\n");
 printf("         -r maximal search radius (100..500000,default 100000)\n");
 printf("         -o output format csv or sql (default csv)\n");
 printf("         -c correct recognizable feet elevations to meter if set to 1 (default 0)\n");
 printf("         -mi miniso minimal distance for screening (default 100)\n");
 printf("         -md mindiff tolerance between DEM and OSM ele (default 40)\n");
 printf("         -i1 min isolation range (below i2 s.rad.) (default 0, means all entries; if set, 2, 3 or 4 groups: i1-radius and radius+, or i1-i2, i2-radius and radius+, or ...\n");
 printf("         -i2 min isolation range (below max.s.rad.) (default 100000)\n");
 printf("         -i3 min isolation range (below max.s.rad.) (default 100000)\n");
 printf("demfile is a geotiff (1 band int16, SRS=EPSG 4326)\n");
 printf("stdin   is a csv with \"id;lon;lat;ele\" where \n");
 printf("         ele is integer, float or empty (not mixed like \"1234m\")\n");
 printf("stdout  is a csv with \"id;lon;lat;isolation\", where\n");
 printf("         isolation is integer [0..radius]\n");
 printf("More documentation you will find in the OSM-Wiki:\n\n");
}

int compare_function_lat(const void *a,const void *b) {
/* helper function for sorting peaks by latitude */
 struct list_peak *x = (struct list_peak *) a;
 struct list_peak *y = (struct list_peak *) b;
 if(y->lat < x->lat) return 1;
 if(y->lat > x->lat) return -1;
 return 0;
}

int compare_function_quot(const void *a,const void *b) {
/* helper function for sorting peaks by quotient */
 struct list_peak *x = (struct list_peak *) a;
 struct list_peak *y = (struct list_peak *) b;
 if(y->quot > x->quot) return 1;
 if(y->quot < x->quot) return -1;
 return 0;
}

double checkele(long long int id, char *elestring){
/* sanitize ele: substitute "," by ".", ignore "m", convert "ft" to meter, check if ele >> Mount Everest */
/* returns ele or OSMNOELE                                                                                 */
 char   *textrest;
 double ele;
 int    i,eleok;

 if(debuglevel>2){printf(" --- checkele:%lld %s\n",id,elestring);}
 
 ele=OSMNOELE;eleok=1;
 if(elestring[0]!='\0'){
  i=0;while(elestring[i]){if(elestring[i]==','){elestring[i]='.';}i++;}
  ele=strtod(elestring,&textrest);
  if(textrest[0]!='\0'){
   eleok=0;
   if(((strstr(textrest,"m"))||(strstr(textrest,"M")))&&(debuglevel>0)){eleok=1;printf(" - %lld has m letters in:%s\n",id,elestring);}
   if (convertFt2Meter){
	if(((strstr(textrest,"f"))||(strstr(textrest,"F"))||(strstr(textrest,"'")))&&(debuglevel>0)){ele=ele*0.3048;eleok=1;printf(" - %lld has ft letters in:%s, convert\n",id,elestring);}
   }else{eleok=1;if(((strstr(textrest,"f"))||(strstr(textrest,"F"))||(strstr(textrest,"'")))&&(debuglevel>0)){;printf(" - %lld has ft letters in:%s no convert\n",id,elestring);}}
  }
  if((!eleok)||(ele>9000.0)||(ele<-12000.0)){ele=OSMNOELE;}
 }else{ele=OSMNOELE;}
 return ele;
}


double get_height(double lon,double lat,double *adfGeoTransform,GDALRasterBandH hBand,long int xsize,long int ysize){
/* get height at lon/lat with elevation data from hBand (parameters in adfGeoTransform) */

 long int   xpx,ypx;
 double     h;
 int16_t    *area2x2;

 /* calculate upper left pixel next to lon/lat  */

 xpx=(long int)floor((lon-adfGeoTransform[0])/adfGeoTransform[1]);
 ypx=(long int)floor((lat-adfGeoTransform[3])/adfGeoTransform[5]);
    
/* read a 2x2 area from demfile at this pixel, so getting all 4 pixel next to lon/lat  */
/* take the max. height from these 4 values                                            */
 
 area2x2=malloc(4*sizeof(int16_t));
 GDALRasterIO(hBand,GF_Read,xpx,ypx,2,2,area2x2,2,2,GDT_Int16,0,0);
 if(debuglevel>3){printf(" ---- Getting x/y:%ld %ld at %lf %lf h=%d %d %d %d\n",xpx,ypx,adfGeoTransform[0]+xpx*adfGeoTransform[1],adfGeoTransform[3]+ypx*adfGeoTransform[5],area2x2[0],area2x2[1],area2x2[2],area2x2[3]);}
 h=area2x2[0];
 if(area2x2[1]>h){h=area2x2[1];}
 if(area2x2[2]>h){h=area2x2[2];}
 if(area2x2[3]>h){h=area2x2[3];}
 free(area2x2);
 if(h<-32000){h=0.0;}
 return h;
}


void get_isolation_by_ele(struct list_peak *peak,long int numpeaks,double radius){
/* gets a list of peaks, calculate distance of each peak to each other in area of radius, returns nothing */

 long int i,j;
 double   ccos,d,dx,dy;
    
 
 for(i=0;i<numpeaks;i++){
  ccos=cos(peak[i].lat*PI/180);
  if(debuglevel>2){printf(" --- calc dist for peak %lld\n",peak[i].id);}
  for(j=i+1;j<numpeaks;j++){
   dy=(peak[j].lat-peak[i].lat)*40000000/360;
   if(debuglevel>3){printf(" ---- compare with other peak %lld\n",peak[j].id);}
   if(dy<radius){
    dx=(peak[j].lon-peak[i].lon)*ccos*40000000/360;
    d=sqrt(dx*dx+dy*dy);
    if(d>radius){d=radius;}
    if(debuglevel>3){printf(" ---- compare with other peak %lld, d=%f\n",peak[j].id,d);}
    if((d < peak[j].isolation)&&(peak[i].ele > peak[j].ele)){
     peak[j].isolation=d;
     peak[j].heigherpeak_id=peak[i].id;
     peak[j].heigherpoint_lon=peak[i].lon;
     peak[j].heigherpoint_lat=peak[i].lat;
     if(debuglevel>3){printf(" ---- assign isolation to other peak %lld, d=%f\n",peak[j].id,d);}
    }
    if((d < peak[i].isolation)     &&(peak[i].ele < peak[j].ele)){
     peak[i].isolation=d;
     peak[i].heigherpeak_id=peak[j].id;
     peak[i].heigherpoint_lon=peak[j].lon;
     peak[i].heigherpoint_lat=peak[j].lat;
     if(debuglevel>2){printf(" ---- found higher peak %lld for peak %lld, at %f/%f d=%f\n",peak[j].id,peak[i].id,peak[j].lat,peak[j].lon,d);}
    }
   }else{
    j=numpeaks+1;
   }
  }
 }
}


void get_isolation_by_DEM(struct list_peak *peak,long int numpeaks,double radius,double *adfGeoTransform,GDALDatasetH hDataset,GDALRasterBandH hBand,long int xsize,long int ysize){
/* gets a list of peaks, calculate isolation of each peak, returns nothing */

 long int  i,up,le,ri,dw,w,h,maxw,maxh,x,y,n,h_ele;
 double    d,r,rx,ry;
 double    ccos;
 int16_t   *DEMarea;
 
 n=0;
 maxw=GDALGetRasterXSize(hDataset);
 maxh=GDALGetRasterYSize(hDataset);
 
 for(i=0;i<numpeaks;i++){
  n++;
  if((n%MODDEM)==0){fprintf(stderr,"%.0lf sec - DEM # %ld %lld\n",difftime(time(NULL),starttime),n,peak[i].id);
    if(debuglevel>3){printf("%.0lf sec - DEM # %ld %lld\n",difftime(time(NULL),starttime),n,peak[i].id);fflush(stdout);}}

  r=peak[i].isolation;
  if(r<miniso){r=miniso;}
  ccos=cos(peak[i].lat*PI/180);
 
/* calc up/le and dw/ri corner of an rectangular area with diameter 2r around each peak */ 
  
  le=(long int)floor((peak[i].lon-r*(1/ccos)*360/40000000-adfGeoTransform[0])/adfGeoTransform[1]);
  ri=(long int) ceil((peak[i].lon+r*(1/ccos)*360/40000000-adfGeoTransform[0])/adfGeoTransform[1]);
  up=(long int)floor((peak[i].lat+r*360/40000000-adfGeoTransform[3])/adfGeoTransform[5]);
  dw=(long int) ceil((peak[i].lat-r*360/40000000-adfGeoTransform[3])/adfGeoTransform[5]); 
  if(le<0){le=0;}
  if(up<0){up=0;}
  if(ri>=maxw){ri=maxw-1;}
  if(dw>=maxh){dw=maxh-1;} 
  w=ri-le;h=dw-up;
  
/* get all DEM values in this area */  
  
  DEMarea=malloc((w)*(h)*sizeof(int16_t));
  GDALRasterIO(hBand,GF_Read,le,up,w,h,DEMarea,w,h,GDT_Int16,0,0); 
  if(debuglevel>2){printf(" --- calc isolation for peak %lld iso=%.7lf (%ld %ld points) le=%ld ri=%ld  dw=%ld up=%ld\n",peak[i].id,r,w,h,le,ri,dw,up);}

/* search for DEM values higher than that peak */

  for(y=up;y<dw;y++){
   for(x=le;x<ri;x++){h_ele=DEMarea[(x-le)+(y-up)*w];
    if(h_ele>peak[i].ele+mindiff){ 
     rx=(peak[i].lon-(adfGeoTransform[0]+adfGeoTransform[1]*x))*40000000/360*ccos;
     ry=(peak[i].lat-(adfGeoTransform[3]+adfGeoTransform[5]*y))*40000000/360;
     d=sqrt(rx*rx+ry*ry);
     if(debuglevel>3){printf(" ---- testing DEM at %ld %ld h=%ld d=%f lat=%f lon=%f\n",x-le,y-up,h_ele,d,adfGeoTransform[3]+adfGeoTransform[5]*y,adfGeoTransform[0]+adfGeoTransform[1]*x);}
     if((d>miniso)&&(d<peak[i].isolation)){
      peak[i].isolation=d;
      peak[i].higherpoint_ele=h_ele;
      peak[i].heigherpoint_lon=adfGeoTransform[0]+adfGeoTransform[1]*x;
      peak[i].heigherpoint_lat=adfGeoTransform[3]+adfGeoTransform[5]*y;
      if(debuglevel>2){printf(" ---- found higher DEM for peak %lld at %ld %ld h=%ld d=%f lat/lon= %f/%f\n",peak[i].id,x-le,y-up,h_ele,d,peak[i].heigherpoint_lat,peak[i].heigherpoint_lon);}
     }
    }
   }
  }
  free(DEMarea);
 }
}


int main(int argc, char *argv[]){

 GDALDatasetH     hDataset;
 GDALRasterBandH  hBand;
 double           adfGeoTransform[6];
 char             *demfile=NULL;
 char             *outputformat=NULL;
 char             *textrest=NULL;
 char 			  *ret;
 double           radius=100000.0,minisol1=0.0,minisol2=radius,minisol3=radius;
 char             line[100],elestring[100];
 long long int    id;
 long int         n=0,i=0,st_s=0,st_g=0,maxnumpeaks=2000000,numpeaks=0;
 double           lon,lat,eleOSM,eleDEM,diff;
 struct list_peak *peak;

/* Parse command line parameters */

 starttime=time(NULL);
 if(argc<3){printhelp();exit(1);}
 for(i=1;i<=argc-2;i+=2){
  if     (strcmp(argv[i],"-d")==0){debuglevel=atoi(argv[i+1]);}
  else if(strcmp(argv[i],"-f")==0){demfile=argv[i+1];}          
  else if(strcmp(argv[i],"-r")==0){radius=strtod(argv[i+1],&textrest);}
  else if(strcmp(argv[i],"-n")==0){maxnumpeaks=strtod(argv[i+1],&textrest);}
  else if(strcmp(argv[i],"-o")==0){outputformat=argv[i+1];}  
  else if(strcmp(argv[i],"-c")==0){convertFt2Meter=atoi(argv[i+1]);}
  else if(strcmp(argv[i],"-mi")==0){miniso=strtod(argv[i+1],&textrest);}
  else if(strcmp(argv[i],"-md")==0){mindiff=strtod(argv[i+1],&textrest);}
  else if(strcmp(argv[i],"-i1")==0){minisol1=strtod(argv[i+1],&textrest);}
  else if(strcmp(argv[i],"-i2")==0){minisol2=strtod(argv[i+1],&textrest);}
  else if(strcmp(argv[i],"-i3")==0){minisol3=strtod(argv[i+1],&textrest);}
  else if(strcmp(argv[i],"-h")==0){printhelp();exit(0);}          
  else if(strcmp(argv[i],"-?")==0){printhelp();exit(0);}
 }   
 if(!demfile){printhelp();exit(1);}
 if(radius<100){radius=100;}
 if(radius>500000){radius=500000;}

 peak=malloc(sizeof(struct list_peak)*maxnumpeaks); 
 if(!peak){
  fprintf(stderr,"Not enough memory for %ld peaks\n",maxnumpeaks);
  exit(1);
 }

 printf("%.0lf sec - Start # miniso: %ld mindiff: %ld  RADIUS: %f convertFt2Meter: %d",difftime(time(NULL),starttime),miniso,mindiff,radius,convertFt2Meter);fflush(stdout);
 fprintf(stderr,"%.0lf sec - Start # miniso: %ld mindiff: %ld  RADIUS: %f convertFt2Meter: %d",difftime(time(NULL),starttime),miniso,mindiff,radius,convertFt2Meter);fflush(stdout);
 
/* open demfile, get srs, origin, pixel size and pointer to the band 1 */ 

 GDALRegister_GTiff();
 hDataset = GDALOpen(demfile,GA_ReadOnly);
 if( hDataset == NULL ){
  fprintf(stderr,"Cannot open File %s\n",demfile);
  exit(1);
 }
 if( GDALGetGeoTransform( hDataset, adfGeoTransform ) == CE_None ){
  if(debuglevel>0){
   fprintf(stderr,"%.0lf sec - Reading %s %dpx x %dpx x %d band\n",difftime(time(NULL),starttime),demfile,GDALGetRasterXSize(hDataset),GDALGetRasterYSize(hDataset),GDALGetRasterCount(hDataset));
   printf("%.0lf sec - Reading %s %dpx x %dpx x %d band\n",difftime(time(NULL),starttime),demfile,GDALGetRasterXSize(hDataset),GDALGetRasterYSize(hDataset),GDALGetRasterCount(hDataset));fflush(stdout);
   printf( " - Origin =(%.6f,%.6f) ",adfGeoTransform[0],adfGeoTransform[3]);
   printf( " - Pixel Size=(%.6f,%.6f)\n",adfGeoTransform[1],adfGeoTransform[5]);
  }
  hBand=GDALGetRasterBand(hDataset,1);
 }else{
  fprintf(stderr,"Could not fetch transformation parameters from %s\n",demfile);
  exit(1);
 }

/* Read stdin, split each line in id,lon,lat,ele */
 n=0;
 while((fgets(line,100,stdin))&&(n<=maxnumpeaks)){
  if((n%MODINP)==0){fprintf(stderr,"%.0lf sec - stdin # %ld %s",difftime(time(NULL),starttime),n,line);
   if(debuglevel>3){printf("%.0lf sec - stdin # %ld %s",difftime(time(NULL),starttime),n,line);fflush(stdout);}}
  
  if(!strstr(line,";")){i=sscanf(line,"%lld %lf %lf %[^\n]",&id,&lon,&lat,elestring);
  }else{i=sscanf(line,"%lld;%lf;%lf;%[^\n]",&id,&lon,&lat,elestring);}
  if(i==3){elestring[0]='\0';}
  if((i>=3)&&(n>maxnumpeaks)){
    fprintf(stderr,"Too many peaks: %ld\n",n);
    exit(1);
   }
  strtok(elestring,";\n");
  eleOSM=checkele(id,elestring);

/* If ele is not parseable, get it from DEMfile... */
/* Then append peak to array of peaks */ 

  eleDEM=get_height(lon,lat,adfGeoTransform,hBand,GDALGetRasterXSize(hDataset),GDALGetRasterYSize(hDataset));
  if((eleOSM<=OSMNOELE)){
   elestring[0]='\0';
   st_s++;
   if(debuglevel>1){
    printf(" -- Input: %s",line);
    printf(" -- > eleDEM: id=%lld lon=%f lat=%f ele=%f\n",id,lon,lat,eleDEM);
   }

   if(debuglevel>2){printf(" --- insert: n=%ld id=%lld lon=%f lat=%f eleDEM=%f\n",n,id,lon,lat,eleDEM);}
    peak[n].id=id;peak[n].lon=lon;peak[n].lat=lat;peak[n].ele=eleDEM;peak[n].eleDEM=eleDEM;peak[n].eleOSM=eleOSM;peak[n].isolation=radius;peak[n].heigherpoint_lon=peak[n].heigherpoint_lat=0.0;
    strncpy(peak[n].elestring,elestring,99);
  }else{
   if(debuglevel>2){printf(" --- insert: n=%ld id=%lld lon=%f lat=%f eleOSM=%f\n",n,id,lon,lat,eleOSM);}
    peak[n].id=id;peak[n].lon=lon;peak[n].lat=lat;peak[n].ele=eleOSM;peak[n].eleDEM=eleDEM;peak[n].eleOSM=eleOSM;peak[n].isolation=radius;peak[n].heigherpoint_lon=peak[n].heigherpoint_lat=0.0;
    strncpy(peak[n].elestring,elestring,99);

  if(peak[n].eleDEM == 0){peak[n].quot=0;
	}else{peak[n].quot=peak[n].eleOSM/peak[n].eleDEM;}	
  }

  n++;numpeaks++;
  st_g++;
 }	/* while stdin */

 fprintf(stderr,"%.0lf sec - Got %ld peaks, %ld without parseable ele\n",difftime(time(NULL),starttime),st_g,st_s); 
 if(debuglevel>0){printf("%.0lf sec - Got %ld peaks, %ld without parseable ele\n",difftime(time(NULL),starttime),st_g,st_s);fflush(stdout);}
 
/* sorting them by latitude */

 qsort(peak,numpeaks,sizeof(struct list_peak),compare_function_lat);
 fprintf(stderr,"%.0lf sec - Sorted peaks by lat\n",difftime(time(NULL),starttime));
 if(debuglevel>0){printf("%.0lf sec - Sorted peaks\n",difftime(time(NULL),starttime));fflush(stdout);}
 
/* Get "isolation" as distance to the next peak */
 
 get_isolation_by_ele(peak,numpeaks,radius);
 fprintf(stderr,"%.0lf sec - Got isolation step 1 (peak to peak distance)\n",difftime(time(NULL),starttime));
 if(debuglevel>0){printf("%.0lf sec - Got isolation step 1 (peak to peak distance)\n",difftime(time(NULL),starttime));fflush(stdout);}

/* Get isolation as distance to the next higher point in DEM */

 get_isolation_by_DEM(peak,numpeaks,radius,adfGeoTransform,hDataset,hBand,GDALGetRasterXSize(hDataset),GDALGetRasterYSize(hDataset));
 fprintf(stderr,"%.0lf sec - Got isolation step 2 (peak to DEM)\n",difftime(time(NULL),starttime));
 if(debuglevel>0){printf("%.0lf sec - Got isolation step 2 (peak to DEM)\n",difftime(time(NULL),starttime));fflush(stdout);}
  
 qsort(peak,numpeaks,sizeof(struct list_peak),compare_function_quot);
 fprintf(stderr,"%.0lf sec - Sorted peaks by quot\n",difftime(time(NULL),starttime));
 if(debuglevel>0){printf("%.0lf sec - Sorted peaks\n",difftime(time(NULL),starttime));fflush(stdout);}
 
 
/* all peaks */
 printf(" -  === all peaks - miniso: %ld mindiff: %ld ===\n",miniso,mindiff);
 for(n=0;n<numpeaks;n++){
	  if((!outputformat)||(strcmp(outputformat,"csv")==0)){
		  
		if(strstr(peak[n].elestring,";")){fprintf(stderr," - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
		if(strstr(peak[n].elestring,";")){printf(" - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
		sprintf(line,"%.3f",peak[n].quot);
		ret=strchr(line,'.');
		if (ret){*ret=',';}		
		if(peak[n].eleOSM == OSMNOELE){diff=peak[n].eleDEM;
		}else{diff=peak[n].eleOSM-peak[n].eleDEM;}	  
		printf("%lld;%.7lf;%.7lf;%.0f;%s;%.0f;%s;%.0f;%lld;%.7lf;%.7lf;%ld\n",peak[n].id,peak[n].lon,peak[n].lat,peak[n].isolation,peak[n].elestring,peak[n].eleDEM,line,diff,peak[n].heigherpeak_id,peak[n].heigherpoint_lon,peak[n].heigherpoint_lat,peak[n].higherpoint_ele);
	  }	/* if outputformat */
	  else if(strcmp(outputformat,"sql")==0) {printf("update planet_osm_point set otm_isolation='%.0lf' where osm_id=%lld;\n",peak[n].isolation,peak[n].id);}
 } /* for */
 if (minisol1 > 0.0){	/* output in two, three or four groups: minisol1..3 (i to i+1) to radius+ */
	 if (minisol1 < minisol2){
		printf(" -  ===== %.0f <= isolation < %.0f - miniso: %ld mindiff: %ld =====\n",minisol1,minisol2,miniso,mindiff);
		for(n=0;n<numpeaks;n++){	/* minisol1 <= isolation < minisol2 */
		  if((!outputformat)||(strcmp(outputformat,"csv")==0)){
			  
			if(strstr(peak[n].elestring,";")){fprintf(stderr," - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
			if(strstr(peak[n].elestring,";")){printf(" - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
			sprintf(line,"%.3f",peak[n].quot);
			ret=strchr(line,'.');
			if (ret){*ret=',';}
			if(peak[n].eleOSM == OSMNOELE){diff=peak[n].eleDEM;
			}else{diff=peak[n].eleOSM-peak[n].eleDEM;}	  
			if((peak[n].isolation >= minisol1)&&(peak[n].isolation < minisol2)) {printf("%lld;%.7lf;%.7lf;%.0f;%s;%.0f;%s;%.0f;%lld;%.7lf;%.7lf;%ld\n",peak[n].id,peak[n].lon,peak[n].lat,peak[n].isolation,peak[n].elestring,peak[n].eleDEM,line,diff,peak[n].heigherpeak_id,peak[n].heigherpoint_lon,peak[n].heigherpoint_lat,peak[n].higherpoint_ele);}
		  }	/* if outputformat */
		  else if(strcmp(outputformat,"sql")==0) {printf("update planet_osm_point set otm_isolation='%.0lf' where osm_id=%lld;\n",peak[n].isolation,peak[n].id);}
		} /* for */
	 }
	if (minisol2 < minisol3){
		printf(" -  ===== %.0f <= isolation < %.0f - miniso: %ld mindiff: %ld =====\n",minisol2,minisol3,miniso,mindiff);
		for(n=0;n<numpeaks;n++){	/* minisol2 <= isolation < minisol3 */
		  if((!outputformat)||(strcmp(outputformat,"csv")==0)){
			  
			if(strstr(peak[n].elestring,";")){fprintf(stderr," - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
			if(strstr(peak[n].elestring,";")){printf(" - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
			sprintf(line,"%.3f",peak[n].quot);
			ret=strchr(line,'.');
			if (ret){*ret=',';}
			if(peak[n].eleOSM == OSMNOELE){diff=peak[n].eleDEM;
			}else{diff=peak[n].eleOSM-peak[n].eleDEM;}	  
			if((peak[n].isolation >= minisol2)&&(peak[n].isolation < minisol3)) {printf("%lld;%.7lf;%.7lf;%.0f;%s;%.0f;%s;%.0f;%lld;%.7lf;%.7lf;%ld\n",peak[n].id,peak[n].lon,peak[n].lat,peak[n].isolation,peak[n].elestring,peak[n].eleDEM,line,diff,peak[n].heigherpeak_id,peak[n].heigherpoint_lon,peak[n].heigherpoint_lat,peak[n].higherpoint_ele);}
		  }	/* if outputformat */
		  else if(strcmp(outputformat,"sql")==0) {printf("update planet_osm_point set otm_isolation='%.0lf' where osm_id=%lld;\n",peak[n].isolation,peak[n].id);}
		} /* for */
	 }
	if (minisol3 < radius){
		printf(" -  ===== %.0f <= isolation < %.0f - miniso: %ld mindiff: %ld =====\n",minisol3,radius,miniso,mindiff);
		for(n=0;n<numpeaks;n++){	/* minisol3 <= isolation < radius */
		  if((!outputformat)||(strcmp(outputformat,"csv")==0)){
			  
			if(strstr(peak[n].elestring,";")){fprintf(stderr," - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
			if(strstr(peak[n].elestring,";")){printf(" - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
			sprintf(line,"%.3f",peak[n].quot);
			ret=strchr(line,'.');
			if (ret){*ret=',';}
			if(peak[n].eleOSM == OSMNOELE){diff=peak[n].eleDEM;
			}else{diff=peak[n].eleOSM-peak[n].eleDEM;}	  
			if((peak[n].isolation >= minisol3)&&(peak[n].isolation < radius)) {printf("%lld;%.7lf;%.7lf;%.0f;%s;%.0f;%s;%.0f;%lld;%.7lf;%.7lf;%ld\n",peak[n].id,peak[n].lon,peak[n].lat,peak[n].isolation,peak[n].elestring,peak[n].eleDEM,line,diff,peak[n].heigherpeak_id,peak[n].heigherpoint_lon,peak[n].heigherpoint_lat,peak[n].higherpoint_ele);}
		  }	/* if outputformat */
		  else if(strcmp(outputformat,"sql")==0) {printf("update planet_osm_point set otm_isolation='%.0lf' where osm_id=%lld;\n",peak[n].isolation,peak[n].id);}
		} /* for */
	 }
	printf(" -  ===== isolation >= %.0f - miniso: %ld mindiff: %ld =====\n",radius,miniso,mindiff);
	for(n=0;n<numpeaks;n++){	/* isolation >= radius */
	  if((!outputformat)||(strcmp(outputformat,"csv")==0)){
		  
		if(strstr(peak[n].elestring,";")){fprintf(stderr," - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
		if(strstr(peak[n].elestring,";")){printf(" - elestring \";\" %lld;\"%s\";%.0f;%.0f\n",peak[n].id,peak[n].elestring,peak[n].eleDEM,peak[n].eleOSM);} 
		sprintf(line,"%.3f",peak[n].quot);
		ret=strchr(line,'.');
		if (ret){*ret=',';}		
		if(peak[n].eleOSM == OSMNOELE){diff=peak[n].eleDEM;
		}else{diff=peak[n].eleOSM-peak[n].eleDEM;}	  
		if(peak[n].isolation >= radius) {printf("%lld;%.7lf;%.7lf;%.0f;%s;%.0f;%s;%.0f;%lld;%.7lf;%.7lf;%ld\n",peak[n].id,peak[n].lon,peak[n].lat,peak[n].isolation,peak[n].elestring,peak[n].eleDEM,line,diff,peak[n].heigherpeak_id,peak[n].heigherpoint_lon,peak[n].heigherpoint_lat,peak[n].higherpoint_ele);}
	  }	/* if outputformat */
	  else if(strcmp(outputformat,"sql")==0) {printf("update planet_osm_point set otm_isolation='%.0lf' where osm_id=%lld;\n",peak[n].isolation,peak[n].id);}
	} /* for */
 }	/* minisol1>0.0 */
 
 fprintf(stderr,"%.0lf sec - Completed\n",difftime(time(NULL),starttime));
 if(debuglevel>0){printf("%.0lf sec - Completed\n",difftime(time(NULL),starttime));fflush(stdout);}
 exit(0);
}



